@@ -1,24 +1,22 @@
-#ifndef CR_RESTORER_H__
-#define CR_RESTORER_H__
+#ifndef __CR_RESTORER_H__
+#define __CR_RESTORER_H__
 
 #include <signal.h>
 #include <limits.h>
+#include <sys/resource.h>
 
 #include "compiler.h"
-#include "types.h"
+#include "asm/types.h"
+#include "asm/fpu.h"
 #include "image.h"
 #include "lock.h"
 #include "util.h"
 #include "crtools.h"
 
-#include "../protobuf/mm.pb-c.h"
-#include "../protobuf/vma.pb-c.h"
-#include "../protobuf/creds.pb-c.h"
-#include "../protobuf/core.pb-c.h"
-
-#ifndef CONFIG_X86_64
-# error Only x86-64 is supported
-#endif
+#include "protobuf/mm.pb-c.h"
+#include "protobuf/vma.pb-c.h"
+#include "protobuf/creds.pb-c.h"
+#include "protobuf/core.pb-c.h"
 
 struct task_restore_core_args;
 struct thread_restore_args;
@@ -66,7 +64,7 @@ struct thread_restore_args {
 	struct restore_mem_zone		mem_zone;
 
 	int				pid;
-	UserX86RegsEntry		gpregs;
+	UserRegsEntry		        gpregs;
 	u64				clear_tid_addr;
 
 	bool				has_futex;
@@ -79,10 +77,15 @@ struct thread_restore_args {
 	struct rst_sched_param		sp;
 
 	struct task_restore_core_args	*ta;
+
+	bool				has_fpu;
+	fpu_state_t			fpu_state;
+
+	u32				tls;
 } __aligned(sizeof(long));
 
 struct task_restore_core_args {
-	struct thread_restore_args	t;			/* thread group leader */
+	struct thread_restore_args	*t;			/* thread group leader */
 
 	int				fd_exe_link;		/* opened self->exe file */
 	int				logfd;
@@ -112,111 +115,17 @@ struct task_restore_core_args {
 	uint32_t			cap_bnd[CR_CAP_SIZE];
 
 	MmEntry				mm;
-	u64				mm_saved_auxv[AT_VECTOR_SIZE];
+	auxv_t				mm_saved_auxv[AT_VECTOR_SIZE];
 	u32				mm_saved_auxv_size;
 	char				comm[TASK_COMM_LEN];
-	TaskKobjIdsEntry		ids;
 
-	int				*rst_tcp_socks;
+	int				nr_rlim;
+	struct rlimit			rlims[RLIM_NLIMITS];
+
+	struct rst_tcp_sock		*rst_tcp_socks;
 	int				rst_tcp_socks_size;
 } __aligned(sizeof(long));
 
-struct pt_regs {
-	unsigned long	r15;
-	unsigned long	r14;
-	unsigned long	r13;
-	unsigned long	r12;
-	unsigned long	bp;
-	unsigned long	bx;
-
-	unsigned long	r11;
-	unsigned long	r10;
-	unsigned long	r9;
-	unsigned long	r8;
-	unsigned long	ax;
-	unsigned long	cx;
-	unsigned long	dx;
-	unsigned long	si;
-	unsigned long	di;
-	unsigned long	orig_ax;
-
-	unsigned long	ip;
-	unsigned long	cs;
-	unsigned long	flags;
-	unsigned long	sp;
-	unsigned long	ss;
-};
-
-struct rt_sigcontext {
-	unsigned long			r8;
-	unsigned long			r9;
-	unsigned long			r10;
-	unsigned long			r11;
-	unsigned long			r12;
-	unsigned long			r13;
-	unsigned long			r14;
-	unsigned long			r15;
-	unsigned long			rdi;
-	unsigned long			rsi;
-	unsigned long			rbp;
-	unsigned long			rbx;
-	unsigned long			rdx;
-	unsigned long			rax;
-	unsigned long			rcx;
-	unsigned long			rsp;
-	unsigned long			rip;
-	unsigned long			eflags;
-	unsigned short			cs;
-	unsigned short			gs;
-	unsigned short			fs;
-	unsigned short			__pad0;
-	unsigned long			err;
-	unsigned long			trapno;
-	unsigned long			oldmask;
-	unsigned long			cr2;
-	struct user_fpregs_entry	*fpstate;
-	unsigned long			reserved1[8];
-};
-
-#ifndef __ARCH_SI_PREAMBLE_SIZE
-#define __ARCH_SI_PREAMBLE_SIZE	(3 * sizeof(int))
-#endif
-
-#define SI_MAX_SIZE	128
-#ifndef SI_PAD_SIZE
-#define SI_PAD_SIZE	((SI_MAX_SIZE - __ARCH_SI_PREAMBLE_SIZE) / sizeof(int))
-#endif
-
-typedef struct rt_siginfo {
-	int	si_signo;
-	int	si_errno;
-	int	si_code;
-	int	_pad[SI_PAD_SIZE];
-} rt_siginfo_t;
-
-typedef struct rt_sigaltstack {
-	void	*ss_sp;
-	int	ss_flags;
-	size_t	ss_size;
-} rt_stack_t;
-
-struct rt_ucontext {
-	unsigned long		uc_flags;
-	struct rt_ucontext	*uc_link;
-	rt_stack_t		uc_stack;
-	struct rt_sigcontext	uc_mcontext;
-	rt_sigset_t		uc_sigmask;	/* mask last for extensibility */
-};
-
-struct rt_sigframe {
-	char			*pretcode;
-	struct rt_ucontext	uc;
-	struct rt_siginfo	info;
-
-	/* fp state follows here */
-};
-
-
 #define SHMEMS_SIZE	4096
 
 /*
@@ -285,4 +194,4 @@ find_shmem(struct shmems *shmems, unsigned long shmid)
 			(vma_entry_is(vma, VMA_ANON_PRIVATE) || \
 			vma_entry_is(vma, VMA_FILE_PRIVATE)))
 
-#endif /* CR_RESTORER_H__ */
+#endif /* __CR_RESTORER_H__ */
