@@ -12,6 +12,7 @@
 #include "syscall.h"
 #include "namespaces.h"
 #include "sysctl.h"
+#include "ipc_ns.h"
 
 #include "protobuf.h"
 #include "protobuf/ipc-var.pb-c.h"
@@ -68,25 +69,25 @@ static void pr_info_ipc_sem_entry(const IpcSemEntry *sem)
 	print_on_level(LOG_INFO, "nsems: %-10d\n", sem->nsems);
 }
 
-static int dump_ipc_sem_set(int fd, const IpcSemEntry *entry)
+static int dump_ipc_sem_set(int fd, const IpcSemEntry *sem)
 {
 	int ret, size;
 	u16 *values;
 
-	size = sizeof(u16) * entry->nsems;
+	size = sizeof(u16) * sem->nsems;
 	values = xmalloc(size);
 	if (values == NULL) {
 		pr_err("Failed to allocate memory for semaphore set values\n");
 		ret = -ENOMEM;
 		goto out;
 	}
-	ret = semctl(entry->desc->id, 0, GETALL, values);
+	ret = semctl(sem->desc->id, 0, GETALL, values);
 	if (ret < 0) {
 		pr_perror("Failed to get semaphore set values");
 		ret = -errno;
 		goto out;
 	}
-	pr_info_ipc_sem_array(entry->nsems, values);
+	pr_info_ipc_sem_array(sem->nsems, values);
 
 	ret = write_img_buf(fd, values, round_up(size, sizeof(u64)));
 	if (ret < 0) {
@@ -155,7 +156,7 @@ static int dump_ipc_sem(int fd)
 
 static void pr_info_ipc_msg(int nr, const IpcMsg *msg)
 {
-	print_on_level(LOG_INFO, "  %-5d: type: %-20ld size: %-10d\n",
+	print_on_level(LOG_INFO, "  %-5d: type: %-20"PRId64" size: %-10d\n",
 		       nr++, msg->mtype, msg->msize);
 }
 
@@ -166,7 +167,7 @@ static void pr_info_ipc_msg_entry(const IpcMsgEntry *msg)
 		       msg->qbytes, msg->qnum);
 }
 
-static int dump_ipc_msg_queue_messages(int fd, const IpcMsgEntry *entry,
+static int dump_ipc_msg_queue_messages(int fd, const IpcMsgEntry *msq,
 				       unsigned int msg_nr)
 {
 	struct msgbuf *message = NULL;
@@ -193,7 +194,7 @@ static int dump_ipc_msg_queue_messages(int fd, const IpcMsgEntry *entry,
 	for (msg_cnt = 0; msg_cnt < msg_nr; msg_cnt++) {
 		IpcMsg msg = IPC_MSG__INIT;
 
-		ret = msgrcv(entry->desc->id, message, msgmax, msg_cnt, IPC_NOWAIT | MSG_COPY);
+		ret = msgrcv(msq->desc->id, message, msgmax, msg_cnt, IPC_NOWAIT | MSG_COPY);
 		if (ret < 0) {
 			pr_perror("Failed to copy IPC message");
 			goto err;
@@ -280,7 +281,7 @@ static int dump_ipc_msg(int fd)
 static void pr_info_ipc_shm(const IpcShmEntry *shm)
 {
 	pr_ipc_desc_entry(LOG_INFO, shm->desc);
-	print_on_level(LOG_INFO, "size: %-10lu\n", shm->size);
+	print_on_level(LOG_INFO, "size: %-10"PRIu64"\n", shm->size);
 }
 
 static int ipc_sysctl_req(IpcVarEntry *e, int op)
@@ -436,7 +437,7 @@ int dump_ipc_ns(int ns_pid, const struct cr_fdset *fdset)
 {
 	int ret;
 
-	ret = switch_ns(ns_pid, CLONE_NEWIPC, "ipc", NULL);
+	ret = switch_ns(ns_pid, &ipc_ns_desc, NULL);
 	if (ret < 0)
 		return ret;
 
@@ -509,12 +510,12 @@ void show_ipc_var(int fd, struct cr_options *o)
 	pb_show_vertical(fd, PB_IPCNS_VAR);
 }
 
-static int prepare_ipc_sem_values(int fd, const IpcSemEntry *entry)
+static int prepare_ipc_sem_values(int fd, const IpcSemEntry *sem)
 {
 	int ret, size;
 	u16 *values;
 
-	size = sizeof(u16) * entry->nsems;
+	size = sizeof(u16) * sem->nsems;
 	values = xmalloc(size);
 	if (values == NULL) {
 		pr_err("Failed to allocate memory for semaphores set values\n");
@@ -529,9 +530,9 @@ static int prepare_ipc_sem_values(int fd, const IpcSemEntry *entry)
 		goto out;
 	}
 
-	pr_info_ipc_sem_array(entry->nsems, values);
+	pr_info_ipc_sem_array(sem->nsems, values);
 
-	ret = semctl(entry->desc->id, 0, SETALL, values);
+	ret = semctl(sem->desc->id, 0, SETALL, values);
 	if (ret < 0) {
 		pr_perror("Failed to set semaphores set values");
 		ret = -errno;
@@ -541,13 +542,14 @@ out:
 	return ret;
 }
 
-static int prepare_ipc_sem_desc(int fd, const IpcSemEntry *entry)
+static int prepare_ipc_sem_desc(int fd, const IpcSemEntry *sem)
 {
 	int ret, id;
 	struct sysctl_req req[] = {
-		{ "kernel/sem_next_id", &entry->desc->id, CTL_U32 },
+		{ "kernel/sem_next_id", &sem->desc->id, CTL_U32 },
 		{ },
 	};
+	struct semid_ds semid;
 
 	ret = sysctl_op(req, CTL_WRITE);
 	if (ret < 0) {
@@ -555,20 +557,35 @@ static int prepare_ipc_sem_desc(int fd, const IpcSemEntry *entry)
 		return ret;
 	}
 
-	id = semget(entry->desc->key, entry->nsems,
-		     entry->desc->mode | IPC_CREAT | IPC_EXCL);
+	id = semget(sem->desc->key, sem->nsems,
+		     sem->desc->mode | IPC_CREAT | IPC_EXCL);
 	if (id == -1) {
 		pr_perror("Failed to create sem set");
 		return -errno;
 	}
 
-	if (id != entry->desc->id) {
+	if (id != sem->desc->id) {
 		pr_err("Failed to restore sem id (%d instead of %d)\n",
-							id, entry->desc->id);
+							id, sem->desc->id);
 		return -EFAULT;
 	}
 
-	ret = prepare_ipc_sem_values(fd, entry);
+	ret = semctl(id, sem->nsems, IPC_STAT, &semid);
+	if (ret == -1) {
+		pr_err("Failed to get sem stat structure\n");
+		return -EFAULT;
+	}
+
+	semid.sem_perm.uid = sem->desc->uid;
+	semid.sem_perm.gid = sem->desc->gid;
+
+	ret = semctl(id, sem->nsems, IPC_SET, &semid);
+	if (ret == -1) {
+		pr_err("Failed to set sem uid and gid\n");
+		return -EFAULT;
+	}
+
+	ret = prepare_ipc_sem_values(fd, sem);
 	if (ret < 0) {
 		pr_err("Failed to update sem pages\n");
 		return ret;
@@ -586,9 +603,9 @@ static int prepare_ipc_sem(int pid)
 		return -1;
 
 	while (1) {
-		IpcSemEntry *entry;
+		IpcSemEntry *sem;
 
-		ret = pb_read_one_eof(fd, &entry, PB_IPCNS_SEM);
+		ret = pb_read_one_eof(fd, &sem, PB_IPCNS_SEM);
 		if (ret < 0) {
 			ret = -EIO;
 			goto err;
@@ -596,10 +613,10 @@ static int prepare_ipc_sem(int pid)
 		if (ret == 0)
 			break;
 
-		pr_info_ipc_sem_entry(entry);
+		pr_info_ipc_sem_entry(sem);
 
-		ret = prepare_ipc_sem_desc(fd, entry);
-		ipc_sem_entry__free_unpacked(entry, NULL);
+		ret = prepare_ipc_sem_desc(fd, sem);
+		ipc_sem_entry__free_unpacked(sem, NULL);
 
 		if (ret < 0) {
 			pr_err("Failed to prepare semaphores set\n");
@@ -613,13 +630,13 @@ err:
 	return ret;
 }
 
-static int prepare_ipc_msg_queue_messages(int fd, const IpcMsgEntry *entry)
+static int prepare_ipc_msg_queue_messages(int fd, const IpcMsgEntry *msq)
 {
 	IpcMsg *msg = NULL;
 	int msg_nr = 0;
 	int ret = 0;
 
-	while (msg_nr < entry->qnum) {
+	while (msg_nr < msq->qnum) {
 		struct msgbuf {
 			long mtype;
 			char mtext[MSGMAX];
@@ -645,7 +662,7 @@ static int prepare_ipc_msg_queue_messages(int fd, const IpcMsgEntry *entry)
 		}
 
 		data.mtype = msg->mtype;
-		ret = msgsnd(entry->desc->id, &data, msg->msize, IPC_NOWAIT);
+		ret = msgsnd(msq->desc->id, &data, msg->msize, IPC_NOWAIT);
 		if (ret < 0) {
 			pr_perror("Failed to send IPC message");
 			ret = -errno;
@@ -659,13 +676,14 @@ static int prepare_ipc_msg_queue_messages(int fd, const IpcMsgEntry *entry)
 	return ret;
 }
 
-static int prepare_ipc_msg_queue(int fd, const IpcMsgEntry *entry)
+static int prepare_ipc_msg_queue(int fd, const IpcMsgEntry *msq)
 {
 	int ret, id;
 	struct sysctl_req req[] = {
-		{ "kernel/msg_next_id", &entry->desc->id, CTL_U32 },
+		{ "kernel/msg_next_id", &msq->desc->id, CTL_U32 },
 		{ },
 	};
+	struct msqid_ds msqid;
 
 	ret = sysctl_op(req, CTL_WRITE);
 	if (ret < 0) {
@@ -673,19 +691,34 @@ static int prepare_ipc_msg_queue(int fd, const IpcMsgEntry *entry)
 		return ret;
 	}
 
-	id = msgget(entry->desc->key, entry->desc->mode | IPC_CREAT | IPC_EXCL);
+	id = msgget(msq->desc->key, msq->desc->mode | IPC_CREAT | IPC_EXCL);
 	if (id == -1) {
 		pr_perror("Failed to create msg set");
 		return -errno;
 	}
 
-	if (id != entry->desc->id) {
+	if (id != msq->desc->id) {
 		pr_err("Failed to restore msg id (%d instead of %d)\n",
-							id, entry->desc->id);
+							id, msq->desc->id);
 		return -EFAULT;
 	}
 
-	ret = prepare_ipc_msg_queue_messages(fd, entry);
+	ret = msgctl(id, IPC_STAT, &msqid);
+	if (ret == -1) {
+		pr_err("Failed to get msq stat structure\n");
+		return -EFAULT;
+	}
+
+	msqid.msg_perm.uid = msq->desc->uid;
+	msqid.msg_perm.gid = msq->desc->gid;
+
+	ret = msgctl(id, IPC_SET, &msqid);
+	if (ret == -1) {
+		pr_err("Failed to set msq queue uid and gid\n");
+		return -EFAULT;
+	}
+
+	ret = prepare_ipc_msg_queue_messages(fd, msq);
 	if (ret < 0) {
 		pr_err("Failed to update message queue messages\n");
 		return ret;
@@ -703,9 +736,9 @@ static int prepare_ipc_msg(int pid)
 		return -1;
 
 	while (1) {
-		IpcMsgEntry *entry;
+		IpcMsgEntry *msq;
 
-		ret = pb_read_one_eof(fd, &entry, PB_IPCNS_MSG_ENT);
+		ret = pb_read_one_eof(fd, &msq, PB_IPCNS_MSG_ENT);
 		if (ret < 0) {
 			pr_err("Failed to read IPC messages queue\n");
 			ret = -EIO;
@@ -714,10 +747,10 @@ static int prepare_ipc_msg(int pid)
 		if (ret == 0)
 			break;
 
-		pr_info_ipc_msg_entry(entry);
+		pr_info_ipc_msg_entry(msq);
 
-		ret = prepare_ipc_msg_queue(fd, entry);
-		ipc_msg_entry__free_unpacked(entry, NULL);
+		ret = prepare_ipc_msg_queue(fd, msq);
+		ipc_msg_entry__free_unpacked(msq, NULL);
 
 		if (ret < 0) {
 			pr_err("Failed to prepare messages queue\n");
@@ -759,6 +792,7 @@ static int prepare_ipc_shm_seg(int fd, const IpcShmEntry *shm)
 		{ "kernel/shm_next_id", &shm->desc->id, CTL_U32 },
 		{ },
 	};
+	struct shmid_ds shmid;
 
 	ret = sysctl_op(req, CTL_WRITE);
 	if (ret < 0) {
@@ -779,6 +813,21 @@ static int prepare_ipc_shm_seg(int fd, const IpcShmEntry *shm)
 		return -EFAULT;
 	}
 
+	ret = shmctl(id, IPC_STAT, &shmid);
+	if (ret == -1) {
+		pr_err("Failed to get shm stat structure\n");
+		return -EFAULT;
+	}
+
+	shmid.shm_perm.uid = shm->desc->uid;
+	shmid.shm_perm.gid = shm->desc->gid;
+
+	ret = shmctl(id, IPC_SET, &shmid);
+	if (ret == -1) {
+		pr_err("Failed to set shm uid and gid\n");
+		return -EFAULT;
+	}
+
 	ret = prepare_ipc_shm_pages(fd, shm);
 	if (ret < 0) {
 		pr_err("Failed to update shm pages\n");
@@ -873,3 +922,8 @@ int prepare_ipc_ns(int pid)
 		return ret;
 	return 0;
 }
+
+struct ns_desc ipc_ns_desc = {
+	.cflag = CLONE_NEWIPC,
+	.str = "ipc",
+};
